import supervisely.nn.checkpoints.yolov5 as yolov5
import supervisely.nn.checkpoints.yolov5_v2 as yolov5_v2
import supervisely.nn.checkpoints.yolov8 as yolov8
import supervisely.nn.checkpoints.detectron2 as detectron2
import supervisely.nn.checkpoints.mmdetection as mmdetection
import supervisely.nn.checkpoints.mmdetection3 as mmdetection3
import supervisely.nn.checkpoints.mmsegmentation as mmsegmentation
import supervisely.nn.checkpoints.mmclassification as mmclassification
import supervisely.nn.checkpoints.ritm as ritm
import supervisely.nn.checkpoints.unet as unet
import supervisely.nn.checkpoints.hrda as hrda
